"""
DynamoDB 를 사용하기 위한 인터페이스
"""
import time
import contextlib
from sawsi.aws.locking import wrapper
from sawsi.aws import shared


class LockingAPI:
    """
    DynamoDB를 이용한 로킹 서비스
    """
    def __init__(self, table_name, credentials=None, region=shared.DEFAULT_REGION):
        """
        :param table_name: "테이블 이름"
        :param credentials: {
            "aws_access_key_id": "str",
            "aws_secret_access_key": "str",
            "region_name": "str",
            "profile_name": "str"
        }
        """
        self.boto3_session = shared.get_boto_session(credentials)
        self.cache = {}
        self.table_name = table_name
        self.dynamoFDB = wrapper.DynamoFDB(self.boto3_session, table_name, region)

    def init_table(self):
        # 테이블을 초기화합니다.
        self.dynamoFDB.init_fdb_table()

    @contextlib.contextmanager
    def lock(self, object_key, max_alive_ts=30, max_retries=3, wait_time=2):
        """
        Lock을 획득하고 해제하는 컨텍스트 매니저.
        """
        acquired = self.acquire_lock(object_key, max_alive_ts, max_retries, wait_time)
        try:
            if not acquired:
                raise RuntimeError("Failed to acquire lock")
            yield  # with 블록 내부의 코드가 실행됩니다.
        finally:
            if acquired:
                self.release_lock(object_key)

    def acquire_lock(self, object_key, max_alive_ts=30, max_retries=3, wait_time=2):
        """
        Lock 을 얻기 위해 시도 합니다.
        :param object_key: 락 대상 객체 키
        :param max_alive_ts: 락이 살아 있을 수 있는 최대 시간 (초)
        :param max_retries: 락이 걸려 있을시 재시도 횟수
        :param wait_time: 기본 재시도 텀 (초)
        :return:
        """
        for i in range(max_retries):
            try:
                # 락을 획득하려고 시도합니다.
                self.dynamoFDB.put_item({
                    '_pk': object_key,
                    '_sk': '_',
                    'ttl': int(time.time() + max_alive_ts)
                }, can_overwrite=False)
                return True
            except Exception as e:
                print(f"Lock acquisition attempt {i + 1} failed. Retrying...")
                time.sleep(wait_time)
                wait_time *= 2  # 대기 시간을 두배로 늘립니다.
        # 모든 재시도가 실패했을 경우 False를 반환합니다.
        return False

    def release_lock(self, object_key):
        # 먼저 일관성 읽기로 아이템 존재를 보장한 후에 제거
        self.dynamoFDB.get_item(object_key, '_', consistent_read=True)
        return self.dynamoFDB.delete_item(object_key, '_')



if __name__ == '__main__':
    lapi = LockingAPI('locking')
    OBJECT_KEY = 'K'
    if lapi.acquire_lock(OBJECT_KEY):
        print("Lock acquired!")
        # ... 다음 작업을 수행 ...
    else:
        print("Failed to acquire lock after maximum retries.")

    lapi.release_lock(OBJECT_KEY)




import json
import io
import time

from sawsi.aws import shared
from sawsi.aws.s3 import wrapper as s3_wrapper
from sawsi.aws.iam import wrapper as iam_wrapper
from sawsi.aws.firehose import wrapper as firehose_wrapper
from json import JSONDecodeError


def _aws_jsons_to_json(content):
    # AWS 의 json 델리미터 (구분자) 가 없어서, 아래같은 특수한 방식으로 가져와야 함.
    decoder = json.JSONDecoder()

    content_length = len(content)
    decode_index = 0
    objs = []
    while decode_index < content_length:
        try:
            obj, decode_index = decoder.raw_decode(content, decode_index)
            objs.append(obj)
        except JSONDecodeError as e:
            print("JSONDecodeError:", e)
            # Scan forward and keep trying to decode
            decode_index += 1
    return objs


class Firehose:
    """
    Firehose API
    """
    def __init__(self, delivery_stream_name, bucket_name, object_key_prefix, error_prefix='error', credentials=None, region=shared.DEFAULT_REGION):
        """
        :param delivery_stream_name:
        :param bucket_name: S3는 재활용할수도 있습니다.
        :param object_key_prefix:
        :param error_prefix:
        :param credentials: {
            "aws_access_key_id": "str",
            "aws_secret_access_key": "str",
            "region_name": "str",
            "profile_name": "str",
        }
        """
        self.boto3_session = shared.get_boto_session(credentials)
        self.cache = {}
        self.delivery_stream_name = delivery_stream_name
        self.bucket_name = bucket_name
        self.object_key_prefix = object_key_prefix
        self.error_prefix = error_prefix
        self.s3 = s3_wrapper.S3(self.boto3_session, region=region)
        self.iam = iam_wrapper.IAM(self.boto3_session, region=region)
        self.firehose = firehose_wrapper.Firehose(self.boto3_session, region=region)

    def init(self):
        """
        실제 버킷 생성 및 Firehose 생성, IAM 연결
        :param: object_key_prefix: "ITEM"
        :param: error_prefix: "ERROR"
        :return:
        """
        self.s3.init_bucket(self.bucket_name)

        role_name = f'RoleKinesis-{self.delivery_stream_name}-{self.bucket_name}'
        role_arn = self._create_role_to_access_s3(role_name)
        self._put_role_policy_to_access_s3(role_name)
        retry_count = 0
        while True:
            try:
                time.sleep(4)
                # 캐시 생성시간 때문에 에러가 날수도 있어서
                self.firehose.create_delivery_stream(
                    self.delivery_stream_name, role_arn, self.bucket_name, self.object_key_prefix, self.error_prefix
                )
                break
            except Exception as ex:
                retry_count += 1
                if retry_count > 2:
                    break
                print(ex)

    def _create_role_to_access_s3(self, role_name):
        # Firehose 에서 S3 접속을 위한 롤 생성
        trust_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {
                        "Service": "firehose.amazonaws.com"
                    },
                    "Action": "sts:AssumeRole"
                }
            ]
        }
        try:
            # IAM 역할 생성
            role_response = self.iam.create_role(
                role_name=role_name, trust_policy=trust_policy, description='Role for Kinesis Firehose to access S3'
            )
            role_arn = role_response['Role']['Arn']
            return role_arn
        except self.iam.client.exceptions.EntityAlreadyExistsException:
            # 역할이 이미 있는 경우 해당 Role의 ARN을 가져온다
            role = self.iam.get_role(role_name)
            return role['Role']['Arn']
        except Exception as ex:
            print('에러 발생:', ex)
            raise ex


    def _put_role_policy_to_access_s3(self, role_name):
        # S3 버킷에 대한 인라인 정책 정의
        s3_policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "s3:AbortMultipartUpload",
                        "s3:GetBucketLocation",
                        "s3:GetObject",
                        "s3:ListBucket",
                        "s3:ListBucketMultipartUploads",
                        "s3:PutObject"
                    ],
                    "Resource": [
                        "arn:aws:s3:::{}".format(self.bucket_name),
                        "arn:aws:s3:::{}/*".format(self.bucket_name)
                    ]
                }
            ]
        }
        try:
            # 인라인 정책을 IAM 역할에 연결
            self.iam.put_role_policy(role_name, 'S3AccessPolicy', s3_policy)
        except Exception as ex:
            print(ex)


    def put_record(self, data):
        """
        파이어호스에 Data 를 삽입합니다.
        :param data: str or bytes
        :return:
        """
        response = self.firehose.put_record(
            self.delivery_stream_name, data
        )
        return response

    def list_objects_v2(self, continuation_token, start_after=None, limit=1000):
        """
        버킷에 있는 객체들을 순환
        :param continuation_token:
        :param start_after:
        :param limit:
        :return:
        """
        response = self.s3.list_objects_v2(bucket_name=self.bucket_name,
                                           prefix=self.object_key_prefix,
                                           limit=limit,
                                           start_after=start_after,
                                           continuation_token=continuation_token)
        return response

    def generate_object_keys(self, start_after=None, limit=1000):
        """
        버킷에 있는 객체들을 순차대로 가져옵니다.
        :param prefix:
        :param start_after:
        :param limit:
        :return:
        """
        continuation_token = None
        response = self.list_objects_v2(continuation_token, start_after, limit)
        continuation_token = response.get('NextContinuationToken', None)
        contents = response.get('Contents', [])
        for content in contents:
            yield content
        while continuation_token:
            response = self.list_objects_v2(continuation_token, start_after, limit)
            continuation_token = response.get('NextContinuationToken', None)
            contents = response.get('Contents', [])
            for content in contents:
                yield content

    def generate_log_json_list_and_key(self, key_start_after=None):
        """
        prefix 로 시작하는, start_after key 지점부터 모든
        log 들을 읽고 내부 컨텐츠를 json 으로 변환하여 list 만들어 출력
        :param key_start_after:
        :return:
        """
        for content in self.generate_object_keys(key_start_after):
            last_key = content['Key']
            buffer = io.BytesIO()
            self.s3.client.download_fileobj(self.bucket_name, last_key, buffer)
            buffer.seek(0)
            item_bytes = buffer.read()
            item_string = item_bytes.decode('utf-8')
            objects = _aws_jsons_to_json(item_string)
            yield objects, last_key


if __name__ == '__main__':
    pass



import time
import botocore
import botocore.client
from boto3.dynamodb.conditions import Key, Attr
from boto3.dynamodb.types import TypeDeserializer, TypeSerializer
from concurrent.futures import ThreadPoolExecutor

type_deserializer = TypeDeserializer()
type_serializer = TypeSerializer()


def divide_chunks(l, n):
    """
    입력된 리스트 l 을 n 개씩 갖도록 split
    :param l:
    :param n:
    :return:
    """
    # looping till length l
    if isinstance(l, list):
        for i in range(0, len(l), n):
            yield l[i:i + n]
    elif isinstance(l, dict):
        items = list(l.items())
        for i in range(0, len(items), n):
            chunks = items[i: i+n]
            dic = {}
            for key, value in chunks:
                dic[key] = value
            yield dic


class DynamoFDB:
    def __init__(self, boto3_session, table_name, region='us-west-2'):
        dynamo_config = botocore.client.Config(max_pool_connections=100)
        self.client = boto3_session.client('dynamodb', config=dynamo_config, region_name=region)
        self.resource = boto3_session.resource('dynamodb', config=dynamo_config, region_name=region)
        self.table_cache = {}
        self.table_name = f'{table_name}'

    def init_fdb_table(self):
        self.create_fdb_table(self.table_name)
        # self.create_fdb_partition_index(self.table_name)

    def create_fdb_table(self, table_name):
        """
        fdb 용 테이블 생성문입니다.
        :param table_name:
        :return:
        """
        try:
            response = self.client.create_table(
                AttributeDefinitions=[
                    {
                        'AttributeName': '_pk',
                        'AttributeType': 'S'
                    }, {
                        'AttributeName': '_sk',
                        'AttributeType': 'S'
                    }
                ],
                TableName=table_name,
                KeySchema=[
                    {
                        'AttributeName': '_pk',
                        'KeyType': 'HASH'
                    }, {
                        'AttributeName': '_sk',
                        'KeyType': 'RANGE'
                    },
                ],
                BillingMode='PAY_PER_REQUEST',
                StreamSpecification={
                    'StreamEnabled': True,
                    'StreamViewType': 'NEW_AND_OLD_IMAGES'
                },
            )
            print('CREATING FDB TABLE...')
            self.client.get_waiter('table_exists').wait(TableName=table_name)
            return response
        except Exception as ex:
            print(ex)
            return True

    def create_fdb_partition_index(self, table_name, index_name, pk_name, sk_name):
        """
        fdb 용 파티션 쿼리 전용 인덱스 생성문입니다.
        :param table_name:
        :return:
        """
        try:
            response = self.client.update_table(
                AttributeDefinitions=[
                    {
                        'AttributeName': pk_name,
                        'AttributeType': 'S'
                    }, {
                        'AttributeName': sk_name,
                        'AttributeType': 'S'
                    }
                ],
                TableName=table_name,
                GlobalSecondaryIndexUpdates=[
                    {
                        'Create': {
                            'IndexName': index_name,
                            'KeySchema': [
                                {
                                    'AttributeName': pk_name,
                                    'KeyType': 'HASH'
                                }, {
                                    'AttributeName': sk_name,
                                    'KeyType': 'RANGE'
                                },
                            ],
                            'Projection': {
                                'ProjectionType': 'ALL',
                            },
                        },
                    }
                ]
            )
            print('UPDATE FDB TABLE INDEX...')
            return response
        except Exception as ex:
            print(ex)
            return True

    def delete_fdb_table(self):
        try:
            response = self.client.delete_table(
                TableName=self.table_name
            )
            return response
        except BaseException as ex:
            print(ex)
            return None

    def delete_item(self, pk, sk):
        response = self.get_table(self.table_name).delete_item(
            Key={
                '_pk': pk,
                '_sk': sk,
            },
            ReturnValues='ALL_OLD'
        )
        return response

    def get_table(self, table_name):
        # 캐싱된 테이블 객체 반환
        if table_name in self.table_cache:
            table = self.table_cache[table_name]
        else:
            table = self.resource.Table(table_name)
            self.table_cache[table_name] = table
        return table

    def get_item(self, pk, sk, consistent_read=True):
        table = self.get_table(self.table_name)
        key = {
            '_pk': pk,
            '_sk': sk
        }
        response = table.get_item(
            Key=key,
            ConsistentRead=consistent_read
        )
        item = response.get('Item', None)
        return item

    def put_item(self, item, can_overwrite=False):
        table = self.get_table(self.table_name)
        # 디비에 넣기전에 인덱스 타입을 고려하여 데이터를 변경한다.
        if can_overwrite:
            response = table.put_item(
                Item=item,
            )
            return response
        else:
            # 키가 이미 존재했을때 중복 쓰기를 방지하는 조건문 추가
            try:
                response = table.put_item(
                    Item=item,
                    ConditionExpression='attribute_not_exists(#pk) AND attribute_not_exists(#sk)',
                    ExpressionAttributeNames={
                        '#pk': '_pk',
                        '#sk': '_sk',
                    }
                )
                return response
            except self.resource.meta.client.exceptions.ConditionalCheckFailedException:
                raise Exception(f'Item already exist. Check "pk_field" and "sk_field" & "post_sk_fields" combination.')

    @classmethod
    def get_update_expression_attrs_pair(cls, item):
        """
        item 에서 업데이트 표현식과 속성을 튜플로 반환합니다.
        :param item:
        :return: (UpdateExpression, ExpressionAttributeValues)
        """
        expression = 'set'
        attr_names = {}
        attr_values = {}
        for idx, (key, value) in enumerate(item.items()):
            attr_key = '#key{}'.format(idx)
            attr_value = ':val{}'.format(idx)
            expression += ' {}={}'.format(attr_key, attr_value)

            attr_names['{}'.format(attr_key)] = key
            attr_values['{}'.format(attr_value)] = value
            if idx != len(item) - 1:
                expression += ','
        return expression, attr_names, attr_values

    def update_item(self, pk, sk, item):
        """
        아이탬을 업데이트합니다.
        만약 업데이트할 항목이 DB에 없는 경우 에러가 발생합니다.
        :param pk:
        :param sk:
        :param item:
        :return:
        """
        expression, attr_names, attr_values = self.get_update_expression_attrs_pair(item)
        attr_names['#pk'] = '_pk'
        attr_names['#sk'] = '_sk'
        try:
            response = self.get_table(self.table_name).update_item(
                Key={'_pk': pk, '_sk': sk},
                UpdateExpression=expression,
                ExpressionAttributeValues=attr_values,
                ExpressionAttributeNames=attr_names,
                ReturnValues="ALL_NEW",
                ConditionExpression='attribute_exists(#pk) AND attribute_exists(#sk)'
            )
        except Exception as ex:
            raise Exception('Item to update not exist')
        return response

    def batch_put(self, items, can_overwrite=False):
        table = self.get_table(self.table_name)
        if can_overwrite:
            # 중복 가능하면
            overwrite_by_keys = None
        else:
            # 불가능하면
            overwrite_by_keys = ['_pk', '_sk']
        with table.batch_writer(overwrite_by_pkeys=overwrite_by_keys) as batch:
            for item in items:
                batch.put_item(
                    Item=item,
                )
        return True

    def batch_delete(self, pk_sk_pairs):
        """
        :param pk_sk_pairs: [
            {
                'pk': '...',
                'sk': '...'
            }, ...
        ]
        :return:
        """
        table = self.get_table(self.table_name)
        with table.batch_writer() as batch:
            for pk_sk_pair in pk_sk_pairs:
                key = {
                    '_pk': pk_sk_pair['_pk'],
                    '_sk': pk_sk_pair['_sk']
                }
                batch.delete_item(Key=key)
        return True

    def _get_items(self, pk_sk_pairs, consistent_read=False, retry_attempt=0):
        keys = list([{
            '_pk': {'S': pk_sk_pair['_pk']},
            '_sk': {'S': pk_sk_pair['_sk']}
        } for pk_sk_pair in pk_sk_pairs if pk_sk_pair])
        if keys:
            response = self.client.batch_get_item(
                RequestItems={
                    self.table_name: {
                        'Keys': keys,
                        'ConsistentRead': consistent_read
                    }
                }
            )

            items_succeed = response['Responses'][self.table_name]

            # response 제대로 안왔을때 재시도 로직
            unprocessed_keys = response.get('UnprocessedKeys', {}).get(self.table_name, {}).get('Keys', [])
            if unprocessed_keys:
                # Backoff al.
                time.sleep(pow(retry_attempt + 1, 2))
                items_to_extend = self._get_items(unprocessed_keys, consistent_read, retry_attempt + 1)
                items_succeed.extend(items_to_extend)
        else:  # Keys 가 없을시 성공 내역 없음
            items_succeed = []

        for item in items_succeed:
            for key, value in item.items():
                value = type_deserializer.deserialize(value)
                item[key] = value

        return items_succeed

    def get_items(self, pk_sk_pairs, consistent_read=False):
        chunks = list(divide_chunks(pk_sk_pairs, 100))
        items_succeed = []
        futures = []
        # 배치 + 멀티스레드로 가져옵니다.
        with ThreadPoolExecutor(max_workers=len(chunks)) as worker:
            for chunk in chunks:
                futures.append(worker.submit(self._get_items, chunk, consistent_read))
        for future in futures:
            items_succeed.extend(future.result())

        # 요청한 순서대로 정렬합니다.
        items_by_key = {(item.get('_pk', ''), item.get('_sk', '')): item for item in items_succeed}
        sorted_items = []
        for pk_sk in pk_sk_pairs:
            if pk_sk:
                item = items_by_key.get((pk_sk['_pk'], pk_sk['_sk']), None)
                sorted_items.append(item)
            else:
                sorted_items.append(None)
        return sorted_items

    def query_items(self, partition_key_name, partition_key_value,
                    sort_condition, sort_key_name, sort_key_value, sort_key_second_value=None, filters=None,
                    start_key=None, reverse=False, limit=100, consistent_read=False, index_name=None,
                    recursive_filters=None):
        """
        AWS BOTO3 전용으로 쿼리 메서드 랩핑
        :param partition_key_name: 파티션키 속성의 이름
        :param partition_key_value: 파티션키 속성의 값
        :param sort_condition: 소트키 조건
        :param sort_key_name:
        :param sort_key_value:
        :param sort_key_second_value:
        :param filters: [
            {
                'field': '<FIELD>',
                'value': '<VALUE>',
                'condition': 'eq' | 'neq' | 'lte' | 'lt' | 'gte' | 'gt' | 'btw' | 'stw' |
                        'is_in' | 'contains' | 'exist' | 'not_exist'
            }
        ]
        :param start_key:
        :param reverse:
        :param limit:
        :param consistent_read:
        :param index_name:
        :param recursive_filters: {  재귀 필터, or, and 연산을 묶어서 수행 가능합니다.
                'left': {
                    'field': '<FIELD>',
                    'value': '<VALUE>',
                    'condition': 'eq' | 'neq' | 'lte' | 'lt' | 'gte' | 'gt' | 'btw' | 'stw' |
                            'is_in' | 'contains' | 'exist' | 'not_exist'
                },
                'operation': 'and',
                'right': {
                    'left': {
                        'field': '<FIELD>',
                        'value': '<VALUE>',
                        'condition': 'eq' | 'neq' | 'lte' | 'lt' | 'gte' | 'gt' | 'btw' | 'stw' |
                                'is_in' | 'contains' | 'exist' | 'not_exist'
                    },
                    'operation': 'or',
                    'right': {
                        'field': '<FIELD>',
                        'value': '<VALUE>',
                        'condition': 'eq' | 'neq' | 'lte' | 'lt' | 'gte' | 'gt' | 'btw' | 'stw' |
                                'is_in' | 'contains' | 'exist' | 'not_exist'
                    },
                },
            }
        :return:
        """

        table = self.get_table(self.table_name)
        key_expression = Key(partition_key_name).eq(partition_key_value)
        if sort_condition == 'eq':
            key_expression &= Key(sort_key_name).eq(sort_key_value)
        elif sort_condition == 'lte':
            key_expression &= Key(sort_key_name).lte(sort_key_value)
        elif sort_condition == 'lt':
            key_expression &= Key(sort_key_name).lt(sort_key_value)
        elif sort_condition == 'gte':
            key_expression &= Key(sort_key_name).gte(sort_key_value)
        elif sort_condition == 'gt':
            key_expression &= Key(sort_key_name).gt(sort_key_value)
        elif sort_condition == 'btw':
            key_expression &= Key(sort_key_name).between(sort_key_value, sort_key_second_value)
        elif sort_condition == 'stw':
            key_expression &= Key(sort_key_name).begins_with(sort_key_value)
        elif sort_condition is None:
            pass
        else:
            raise Exception('sort_type must be one of [eq, lte, lt, gte, gt, btw, stw]')

        filter_expression = None
        if filters:
            for ft in filters:
                attr_to_add = self.get_attr(ft)
                if filter_expression:
                    filter_expression &= attr_to_add
                else:
                    filter_expression = attr_to_add
        elif recursive_filters:
            # 재귀 필터가 넘어온 경우, 딕셔너리 형태입니다.
            filter_expression = self.get_recursive_filters_expression(recursive_filters)

        args = {
            'Limit': limit,
            'ConsistentRead': consistent_read,
            'KeyConditionExpression': key_expression,
            'ScanIndexForward': not reverse,
        }
        if index_name:
            args['IndexName'] = index_name
            args['ConsistentRead'] = False  # index 사용시 일관된 읽기는 사용 불가
        if filter_expression:
            args['FilterExpression'] = filter_expression
        if start_key:
            args['ExclusiveStartKey'] = start_key

        response = table.query(**args)
        return response

    @classmethod
    def get_recursive_filters_expression(cls, recursive_filters):
        # recursive_filters 로 부터, DynamoDB exp 를 구합니다.
        if not isinstance(recursive_filters, dict):
            raise Exception('<recursive_filters> must be type of <dict>')

        # dict 의 형식을 보고, 연산 결정
        if 'left' in recursive_filters and 'right' in recursive_filters and 'operation' in recursive_filters:
            filter_type = 'tuple'
        elif 'field' in recursive_filters and 'value' in recursive_filters and 'condition' in recursive_filters:
            filter_type = 'value'
        else:
            raise Exception('<recursive_filters> must consist of [<left> & <right> & <operation>] or '
                            '[<field> & <value> & <condition>]')

        # 실제 filter exp 변환
        filter_expression = None
        if filter_type == 'tuple':
            # left, right 타입 있는 경우, 재귀 진행.
            left = cls.get_recursive_filters_expression(recursive_filters['left'])
            right = cls.get_recursive_filters_expression(recursive_filters['right'])
            operation = recursive_filters['operation']
            if operation == 'and':
                filter_expression = left & right
            elif operation == 'or':
                filter_expression = left | right
            return filter_expression
        elif filter_type == 'value':
            # value 인 경우 바로 반환
            return cls.get_attr(recursive_filters)
        else:
            # 있을 수 없는 경우
            raise Exception('<filter_type> Error, must be one of <tuple> or <value>')

    @classmethod
    def get_attr(cls, ft):
        # filter 한개를 DynamoDB 의 ATTR 로 변환합니다.
        field = ft['field']
        value = ft.get('value', None)
        high_value = ft.get('second_value', None)
        cond = ft['condition']

        attr_to_add = Attr(field)
        if cond == 'eq':
            attr_to_add = attr_to_add.eq(value)
        elif cond == 'neq':
            attr_to_add = attr_to_add.ne(value)
        elif cond == 'lte':
            attr_to_add = attr_to_add.lte(value)
        elif cond == 'lt':
            attr_to_add = attr_to_add.lt(value)
        elif cond == 'gte':
            attr_to_add = attr_to_add.gte(value)
        elif cond == 'gt':
            attr_to_add = attr_to_add.gt(value)
        elif cond == 'btw':
            attr_to_add = attr_to_add.between(value, high_value)
        elif cond == 'not_btw':
            attr_to_add = ~attr_to_add.between(value, high_value)
        elif cond == 'stw':
            attr_to_add = attr_to_add.begins_with(value)
        elif cond == 'not_stw':
            attr_to_add = ~attr_to_add.begins_with(value)
        elif cond == 'is_in':
            attr_to_add = attr_to_add.is_in(value)
        elif cond == 'is_not_in':
            attr_to_add = ~attr_to_add.is_in(value)
        elif cond == 'contains':
            attr_to_add = attr_to_add.contains(value)
        elif cond == 'not_contains':
            attr_to_add = ~attr_to_add.contains(value)
        elif cond == 'exist':
            attr_to_add = attr_to_add.exists()
        elif cond == 'not_exist':
            attr_to_add = attr_to_add.not_exists()
        else:
            raise Exception('<condition> parameter must be one of '
                            '[eq, neq, lte, lt, gte, gt, btw, stw, is_in, contains, '
                            'not_btw, not_stw, not_is_in, not_contains, exist, not_exist]')
        return attr_to_add




SK_DIGIT_FIT = 20  # sort key 의 자리 맞춤 총 문자열 개수임.
SK_FLOAT_DEC_FIT = 16  # SK_DIGIT_FIT 에서 소숫점 자리가 차지하는 부분, 데시멀 (10진수)임

# base64 'z' * SK_DIGIT_FIT 로 표현할 수 있는 최대 범위 // 2
HALF_REP_NUMBER_B64 = 664613997892457936451903530140172288
MAX_REP_NUMBER_B64 = HALF_REP_NUMBER_B64 * 2

# 숫자를 base64 형태로 변환하여 표현
BASE64_LETTERS = "+/0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
STR_META_INFO_PARTITION = 'meta-info#partition'
END_NOTATION = '|'


if __name__ == '__main__':
    print(64**20 // 2)


"""
핸들러에서 쓰이는 유틸
에러수집 or payload, user 로 바꾸는 부분
에러 수정 유틸.
"""
import traceback
import json
from sawsi.shared import error_util
from sawsi.shared import dict_util
import time


def get_body(event, context):
    """
    API로부터 요청된 body 를 반환합니다.
    :param event:
    :param context:
    :return: dict
    """
    body = event.get('body', None)
    if body:
        try:
            return json.loads(body)
        except Exception as ex:
            print(ex)
    return body


def get_headers(event, context):
    """
    API로부터 요청된 headers 를 반환합니다.
    :param event:
    :param context:
    :return: dict
    """
    headers = event.get('headers', {})
    return headers


def get_source_ip(event, context):
    """
    요청지의 소스IP를 얻어냅니다.
    :param event:
    :param context:
    :return: 52.411.555.11 (EX)
    """
    request_context = event.get('requestContext', {})
    client_ip = request_context.get('identity', {}).get('sourceIp', None)
    return client_ip


def aws_handler_wrapper(error_receiver=None, content_type='application/json', use_traceback=True):  # 데코레이터 팩토리
    """
    핸들러를 감싸고, 해당 코드에서 에러가 발생하면,
    error_receiver(message) 형식으로 콜백이 이루어집니다.
    외부에서 해당 에러 메시지를 Slack 에 전송하거나 기타 알림을 수행합니다.
    :param error_receiver: def some_function_to_notice_error_message(message: str)
    :param content_type: 'application/json' 등의 반환되는 콘텐츠 타입입니다.
    :param use_traceback: 에러 발생시, 상세 내용 전파 여부입니다.
    :return:
    """
    def _handler_wrapper(func):
        """
        엔트리 핸들러에는 이 데코레이터를 붙여준다.
        반환되는 응답에 코드화를 해줍니다.
        :param func:
        :return:
        """
        def decorator(*args, **kwargs):
            now = time.time()
            print(args, kwargs)
            try:
                result = func(*args, **kwargs)
                if result is None:
                    result = {}
                if 'rslt_cd' not in result:
                    # 없으면 기본으로 삽입
                    result['rslt_cd'] = 'A00000'
                    result['rslt_msg'] = '요청 성공'
                    result['duration'] = time.time() - now
                aws_response = AWSResponse(result, content_type=content_type)
                print(aws_response)
                return aws_response
            except error_util.AppError as ex:
                result = {
                    'rslt_cd': str(ex.code),
                    'rslt_msg': str(ex.message),
                    'duration': time.time() - now
                }
                if use_traceback:
                    result['traceback'] = str(traceback.format_exc())
                if error_receiver:
                    try:
                        message = error_util.make_error_description(args[0])
                        error_receiver(message)
                    except Exception as ex:
                        result['error_receiver_failed_exc'] = str(ex)
                aws_response = AWSResponse(result)
                print(aws_response)
                return aws_response
            except Exception as ex:
                result = {
                    # 알수없는에러
                    'rslt_cd': 'U99999',
                    'rslt_msg': str(ex),
                    'duration': time.time() - now
                }
                if use_traceback:
                    result['traceback'] = str(traceback.format_exc())
                if error_receiver:
                    try:
                        message = error_util.make_error_description(args[0])
                        error_receiver(message)
                    except Exception as ex:
                        result['error_receiver_failed_exc'] = str(ex)
                aws_response = AWSResponse(result)
                print(aws_response)
                return aws_response
        return decorator
    return _handler_wrapper


def _get_headers(content_type='application/json', charset='UTF-8'):
    api_gateway_response_header = {
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Headers': '*',
        'Access-Control-Allow-Credentials': True,
        'Access-Control-Allow-Methods': 'POST,OPTIONS',
        'Content-Type': content_type,
        'charset': charset
    }
    return api_gateway_response_header


class AWSResponse(dict):
    def __init__(self, body, content_type='application/json', status_code=200, charset='UTF-8'):
        headers = _get_headers(content_type, charset=charset)
        self['statusCode'] = status_code
        self['headers'] = headers
        if isinstance(body, dict):
            body = dict_util.convert_decimal_to_number(body)  # 숫자 변경을 위해
            self['body'] = json.dumps(body, default=lambda o: '<not serializable>')
        else:
            self['body'] = body
        self['isBase64Encoded'] = False
